"""Config frame."""

from espmu.lib import hexToBin
from espmu.frame import Frame
from espmu.station import Station
from espmu.parsed import Parsed


class ConfigFrame(Frame):
    """Parses Config Frame (1 or 2).

    Parameters
    ----------
    buf_hex: str
        Config frame as byte array in hex str format.
    debug: bool
        Print debug statements.
    """
    def __init__(self, buf_hex, debug=False):
        super().__init__(buf_hex, debug)
        self.__data = {}
        self.__buf = None

    def finishParsing(self):
        """After first 4 bytes are received, the client reads the
        remaining config frame bytes. This function parses those
        remaining bytes."""
        super().finishParsing()

        self.__buf = Parsed(self.frame)
        self.__buf.pos = self.length

        self.__parse_time_base()
        self.__parse_num_pmu()
        self.__parse_stations()
        self.__parse_datarate()

        self.length = self.__buf.pos

        if self._dbg:
            print("ConfigFrame:", self.__data)

    @property
    def time_base(self):
        """Resolution of FRACSEC."""
        return self.__data['time_base']

    @property
    def num_pmu(self):
        """Number of PMUs sending data."""
        return self.__data['num_pmu']

    @property
    def stations(self):
        """Station for each PMU."""
        return self.__data['stations']

    @property
    def datarate(self):
        """Data rate at which data will be received."""
        return self.__data['datarate']

    def __parse_time_base(self):
        self.__data['time_base'] = TimeBase(self.__buf.next(8))

    def __parse_num_pmu(self):
        self.__data['num_pmu'] = int(self.__buf.next(4), 16)

    def __parse_stations(self):
        self.__data['stations'] = []
        for _ in range(self.num_pmu):
            stn = Station(self.__buf.rest(), debug=self._dbg)
            self.__data['stations'].append(stn)
            self.__buf.pos += stn.pos

    def __parse_datarate(self):
        self.__data['datarate'] = int(self.__buf.next(4), 16)


class TimeBase:
    """The TIME_BASE word."""
    def __init__(self, timeBaseHexStr):
        self.timeBaseHex = timeBaseHexStr
        self.flagsBinStr = hexToBin(timeBaseHexStr, 32)[:8]
        # Not parsed anywhere because so far they aren't being used
        self.baseDecStr = int(timeBaseHexStr[1:], 16)
