"""In this module the base class for frames is defined."""
from datetime import datetime
from espmu.enums import FrameType
from espmu.lib import hexToBin


class Frame:
    """Super class for all C37.118-2005/C37.118-2011 frames.

    Parameters
    ----------
    buf_hex: str
        Bytes from frame (any time) in hex str format.
    debug: bool
        Print debug statements.
    """

    def __init__(self, buf_hex, debug=False):
        self._dbg = debug

        self.length = 0
        self.frame = buf_hex.upper()

        self.__parse_sync()
        self.__parse_framesize()

        self.idcode = None
        self.soc = None
        self.tq = None
        self.fracsec = None
        self.chk = None

        if self._dbg:
            print("FRAMESIZE: ", self.framesize)

    def finishParsing(self):
        """When getting the config frame, the size is unknown.  After
        creating a Frame with the first 4 bytes, the remaining frame
        bytes are read and added to self.frame. Once that is populated
        the remaining fields can be parsed."""
        self.__parse_idcode()
        self.__parse_soc()
        self.__parse_tq()
        self.__parse_fracsec()
        self.__parse_chk()

        if self._dbg:
            print("IDCODE: ", self.idcode)
            print("TQ: ", self.tq)
            print("FRACSEC: ", self.fracsec)
            print("CHK: ", self.chk)

    def _updateLength(self, size_to_add):
        """Keeps track of index for overall frame"""
        self.length = self.length + size_to_add

    def __parse_sync(self):
        """Parse frame synchronization word"""
        self.sync = SYNC(self.frame[:4])
        self._updateLength(4)

    def __parse_framesize(self):
        """Parse frame size"""
        framesize_size = 4
        self.framesize = int(
            self.frame[self.length:self.length+framesize_size],
            16)
        self._updateLength(framesize_size)

    def __parse_idcode(self):
        """Parse data stream ID number"""
        idcode_size = 4
        self.idcode = int(self.frame[self.length:self.length+idcode_size], 16)
        self._updateLength(idcode_size)

    def __parse_soc(self):
        """Parse second-of-century timestamp"""
        soc_size = 8
        self.soc = SOC(self.frame[self.length:self.length+soc_size])
        self._updateLength(soc_size)

    def __parse_tq(self):
        """Parse time guality flag"""
        tq_size = 2
        self.tq = int(self.frame[self.length:self.length+tq_size], 16)
        self._updateLength(tq_size)

    def __parse_fracsec(self):
        """Parse fraction of second and time quality word"""
        fracsec_size = 6
        self.fracsec = int(
            self.frame[self.length:self.length+fracsec_size], 16)
        self._updateLength(fracsec_size)

    def __parse_chk(self):
        """Parse CRC-CCITT word"""
        chk_size = 4
        self.chk = self.frame[-chk_size:]


class SYNC:
    """Class for describing the frame synchronization word.

    Parameters
    ----------
    buf_hex: str
        Sync byte array in hex str format.
    debug: bool
        Print debug statements.
    """

    def __init__(self, buf_hex, debug=False):
        self.syncHex = buf_hex
        self.frameType = None
        self.frameVers = None
        self.__parse_type()
        self.__parse_vers()

        if debug:
            print("Type: ", self.frameType)
            print("Vers: ", self.frameVers)

    def __parse_type(self):
        """Parse frame type"""
        type_bin_str = hexToBin(self.syncHex[2], 3)
        type_num = int(type_bin_str, 2)
        self.frameType = FrameType(type_num).name

    def __parse_vers(self):
        """Parse frame version"""
        vers_bin_str = hexToBin(self.syncHex[3], 4)
        self.frameVers = int(vers_bin_str, 2)


class SOC:
    """Class for second-of-century (SOC) word (32 bit unsigned).

    Parameters
    ----------
    buf_hex: str
        Second-of-century byte array in hex str format.
    debug: bool
        Print debug statements.
    """

    def __init__(self, buf_hex, debug=False):
        self.socHex = buf_hex
        self.secCount = int(buf_hex, 16)
        self.__parse_sec_count()

        if debug:
            print("SOC: ", self.secCount, " - ", self.formatted)

    def __parse_sec_count(self):
        """Parse SOC into UTC timestamp and pretty formatted timestamp."""

        parsed_date = datetime.fromtimestamp(self.secCount)
        self.yyyy = parsed_date.year
        self.mm = parsed_date.month
        self.dd = parsed_date.day
        self.hh = parsed_date.hour
        self.mi = parsed_date.minute
        self.ss = parsed_date.second
        self.ff = 0
        self.formatted = "{:0>4}/{:0>2}/{:0>2} {:0>2}:{:0>2}:{:0>2}".format(
            self.yyyy, self.mm, self.dd, self.hh, self.mi, self.ss
        )
        dt = datetime(self.yyyy, self.mm, self.dd, self.hh, self.mi, self.ss)
        self.utcSec = (dt - datetime(1970, 1, 1)).total_seconds()
