"""Commonly used functions for parsing PMU Frames."""
import codecs
import struct


def hexToBin(hex_str, num_of_bits):
    """Converts hex string to binary.

    Parameters
    ----------
    hex_str: str
        Hex value in string format.
    num_of_bits: int
        Number of bits to convert.

    Returns
    -------
    : str
        Bits representing the hex values.
    """
    return bin(int(hex_str, 16))[2:].zfill(num_of_bits)


def bytesToHexStr(bytes_input):
    """Converts byte array to hex str.

    Parameters
    ----------
    bytes_input: byte-array
        Byte array to convert.

    Returns
    -------
    : str
        Hex string representing bytes_input.
    """

    return codecs.encode(bytes_input, 'hex').decode('ascii')


def doubleToHex(f):
    """Converts double to hex

    Parameters
    ----------
    f: double
        Double value to convert.

    Returns
    -------
    : str
        Hex representation of double value.
    """
    return hex(struct.unpack('!Q', struct.pack('!d', f))[0])


def doubleToHexStr(f):
    """Converts double to hex str.

    Parameters
    ----------
    f: double
        Double value to convert.

    Returns
    -------
    : str
        Hex string representation of double value.
    """
    return hex(struct.unpack('!Q', struct.pack('!d', f))[0])[2:]


def doubleToBytes(f):
    """Converts double to byte array.

    Parameters
    ----------
    f: double
        Double value to convert.

    Returns
    -------
    : bytes
        Byte array representation of double value.
    """
    return struct.pack('d', f)


def bytesToFloat(b):
    """Converts byte array to double.

    Parameters
    ----------
    b: byte-array
        Byte array to convert.

    Returns
    -------
    : tuple
        Float value.
    """
    return struct.unpack('d', b)[0]


def intToBytes(i):
    """Converts unsigned int to byte array.

    Parameters
    ----------
    i: int
        Integer value to convert.

    Returns
    -------
    : bytes
        Byte array representing i.
    """
    return struct.pack('!I', i)


def intToHexStr(i):
    """Converts int to hex

    Parameters
    ----------
    i: int
        Integer value to convert.

    Returns
    -------
    : str
        Hex string representing i.
    """
    return hex(i)[2:]
