"""This module implements the class for high-level interaction with
number of PMUs (stations) whithin PDC."""
from warnings import warn
from espmu import tools
from espmu.client import Client


TIMEOUT_SEC = 5


class Reader:
    """Data reader."""
    def __init__(self):
        """Initialization."""
        self.__idcode = None
        self.__cli = None
        self.__data_on = False
        self.__output_settings = []
        self.__conf_frame = None
        self.__rest = ""

    def connect_to(self, ip_addr, tcp_port, idcode):
        """Connect to PDC or PMU."""
        self.__idcode = idcode

        self.__cli = Client(ip_addr, tcp_port, proto="TCP")
        self.__cli.setTimeout(TIMEOUT_SEC)
        self.__cli.connectToDest()

        tools.turnDataOff(self.__cli, idcode)

        tools.requestConfigFrame2(self.__cli, idcode)
        self.__conf_frame = tools.readConfigFrame2(self.__cli)

        self.__output_settings = [None]*self.__conf_frame.num_pmu

    def config_frame(self):
        """Return config frame."""
        return self.__conf_frame

    def start(self):
        """Start data stream."""
        tools.turnDataOn(self.__cli, self.__idcode)
        self.__rest = ""
        self.__data_on = True

    def stop(self):
        """Stop data stream."""
        tools.turnDataOff(self.__cli, self.__idcode)
        self.__data_on = False

    def disconnect(self):
        """Disconnect from data source."""
        if self.__cli:
            self.__cli.stop()
            self.__cli = None

    def is_data_on(self):
        """Check if data stream is on."""
        return self.__data_on

    def is_time_reliable(self):
        """Check time quality flag and return False if time is not
        reliable, else True."""
        return self.__conf_frame.tq != 15

    def rate(self):
        """Return data rate."""
        return self.__conf_frame.datarate

    def stations(self):
        """Return the names of stations."""
        if not self.__conf_frame:
            return None

        return [strip_name(st_name.stn) for st_name
                in self.__conf_frame.stations]

    def get_data_frames(self):
        """Return the list of data frames."""
        frames, rest = tools.get_data_frames(
            tools.getDataSample(self.__cli),
            self.__conf_frame, self.__rest)

        self.__rest = rest
        return frames

    # Deprecated methods

    def phasors(self, station_ind):
        """Deprecated."""

        warn("phasors() is deprecated, use conf_frame() instead",
             FutureWarning)
        return [strip_name(name) for name
                in self.__conf_frame.stations[station_ind].ph_channels]

    def analogs(self, station_ind):
        """Deprecated."""

        warn("analogs() is deprecated, use conf_frame() instead",
             FutureWarning)
        return [strip_name(name) for name
                in self.__conf_frame.stations[station_ind].an_channels]

    def connect(self, ip_addr, tcp_port, idcode):
        """Deprecated."""

        warn("connect() is deprecated, use connect_to() instead",
             FutureWarning)
        self.connect_to(ip_addr, tcp_port, idcode)
        return True

    def get_full_samples(self, station_ind):
        """Deprecated."""

        warn("get_full_samples() deprecated, use get_data_frames()",
             FutureWarning)

        samples = []
        for data_frame in self.get_data_frames():
            station = data_frame.pmus[station_ind]

            # 0 and 1 - time and freq
            sample = [get_time(data_frame), station.freq]

            # then phasors
            for phasor in station.phasors:
                sample.append((phasor.mag, phasor.rad))

            # and analogs
            for analog in station.analogs:
                sample.append(analog[1])

            samples.append(sample)

        return samples


class PmuStreamDataReader(Reader):
    """Deprecated."""
    def __init__(self):
        super().__init__()
        warn("PmuStreamDataReader deprecated, use Reader")


def strip_name(name):
    """Strip name of channel."""
    return name.split("\x00")[0].strip()


def get_time(data_frame):
    """Get time from data_frame."""

    secs = data_frame.soc.secCount
    msecs = data_frame.fracsec
    return secs + msecs / data_frame.configFrame.time_base.baseDecStr
