"""Configuration of the station."""

from espmu.parsed import Parsed
from espmu.units import Phunit, Anunit, Digunit
from espmu.enums import NumType, PhsrFmt, FundFreq
from espmu.lib import hexToBin


class Station:
    """Class for parsing station information including all PMU
    information. Fields 8-19.

    Parameters
    ----------
    buf_hex: str
        Station fields in hex str format.
    debug: bool
        Print debug info.
    """

    def __init__(self, buf_hex, debug=False):
        self.__buf = Parsed(buf_hex)
        self.__data = {}
        self.__parse_stn()
        self.__parse_idcode_data()
        self.__parse_format()
        self.__parse_phnmr()
        self.__parse_annmr()
        self.__parse_dgnmr()
        self.__parse_phname()
        self.__parse_anname()
        self.__parse_dgname()
        self.__parse_ph_units()
        self.__parse_an_units()
        self.__parse_dg_units()
        self.__parse_fnom()
        self.__parse_cfgcnt()

        if debug:
            print('Station:', self.__data)

    @property
    def stn(self):
        """Station name."""
        return self.__data['stn']

    @property
    def idcode_data(self):
        "Station ID code."
        return self.__data['idcode_data']

    @property
    def freqType(self):
        """Data type for frequency."""
        return self.__data['freq_type']

    @property
    def anlgType(self):
        """Data type for analog."""
        return self.__data['anlg_type']

    @property
    def phsrType(self):
        """Data type for phasor."""
        return self.__data['phsr_type']

    @property
    def phsrFmt(self):
        """Format of phasor."""
        return self.__data['phsr_fmt']

    @property
    def phnmr(self):
        """Number of phasors field."""
        return self.__data['ph_num']

    @property
    def annmr(self):
        """Number of analog values field."""
        return self.__data['an_num']

    @property
    def dgnmr(self):
        """Number of digital values field."""
        return self.__data['dg_num']

    @property
    def ph_channels(self):
        """Phasor channels."""
        return self.__data['ph_channels']

    @property
    def an_channels(self):
        """Analog channels."""
        return self.__data['an_channels']

    @property
    def dg_channels(self):
        """Digital channels."""
        return self.__data['dg_channels']

    @property
    def channels(self):
        """All channels."""
        return [] +\
            self.__data['ph_channels'] +\
            self.__data['an_channels'] +\
            self.__data['dg_channels']

    @property
    def anunits(self):
        """Analog units."""
        return self.__data['an_units']

    @property
    def phunits(self):
        """Phasor units."""
        return self.__data['ph_units']

    @property
    def digunits(self):
        """Digital units."""
        return self.__data['dg_units']

    @property
    def fnom(self):
        """Nominal line frequency code and flags."""
        return self.__data['fnom']

    @property
    def cfgcnt(self):
        """Configuration change count."""
        return self.__data['cfgcnt']

    @property
    def pos(self):
        """Return current position in parsed buffer."""
        return self.__buf.pos

    def __parse_stn(self):
        self.__data['stn'] = bytes.fromhex(self.__buf.next(32)).decode('ascii')

    def __parse_idcode_data(self):
        self.__data['idcode_data'] = int(self.__buf.next(4), 16)

    def __parse_format(self):
        fmts = hexToBin(self.__buf.next(4), 32)[-4:]
        self.__data['freq_type'] = NumType(int(fmts[0], 2)).name
        self.__data['anlg_type'] = NumType(int(fmts[1], 2)).name
        self.__data['phsr_type'] = NumType(int(fmts[2], 2)).name
        self.__data['phsr_fmt'] = PhsrFmt(int(fmts[3], 2)).name

    def __parse_phnmr(self):
        self.__data['ph_num'] = int(self.__buf.next(4), 16)

    def __parse_annmr(self):
        self.__data['an_num'] = int(self.__buf.next(4), 16)

    def __parse_dgnmr(self):
        self.__data['dg_num'] = int(self.__buf.next(4), 16)

    def __parse_phname(self):
        self.__data['ph_channels'] = self.__parse_name(self.phnmr)

    def __parse_anname(self):
        self.__data['an_channels'] = self.__parse_name(self.annmr)

    def __parse_dgname(self):
        self.__data['dg_channels'] = self.__parse_name(16 * self.dgnmr)

    def __parse_name(self, num):
        chs = [None] * num
        for i, _ in enumerate(chs):
            chs[i] = bytes.fromhex(self.__buf.next(32)).decode('ascii')
        return chs

    def __parse_ph_units(self):
        units = [None]*self.phnmr
        for i, _ in enumerate(units):
            units[i] = Phunit(self.__buf.next(8))
        self.__data['ph_units'] = units

    def __parse_an_units(self):
        units = [None]*self.annmr
        for i, _ in enumerate(units):
            units[i] = Anunit(self.__buf.next(8))
        self.__data['an_units'] = units

    def __parse_dg_units(self):
        units = [None]*self.dgnmr
        for i, _ in enumerate(units):
            units[i] = Digunit(self.__buf.next(8))
        self.__data['dg_units'] = units

    def __parse_fnom(self):
        hex_digit = self.__buf.next(4)
        hex_digit_lsb = hexToBin(hex_digit, 8)[7]
        hex_digit_dec = int(hex_digit_lsb, 2)
        self.__data['fnom'] = FundFreq(hex_digit_dec).name

    def __parse_cfgcnt(self):
        """Configuration change count."""
        self.__data['cfgcnt'] = int(self.__buf.next(4), 16)
