"""Tools for common functions relayed to commanding, reading, and
parsing PMU data."""

from espmu.client import Client
from espmu.config_frame import ConfigFrame
from espmu.command_frame import CommandFrame
from espmu.data_frame import DataFrame
from espmu.lib import bytesToHexStr

MAXFRAMESIZE = 65535


def turnDataOff(cli, idcode):
    """Send command to turn off real-time data.

    Parameters
    ----------
    cli: Client
        Client being used to connect to data source.
    idcode: int
        Frame ID of data source.
    """
    cmd_off = CommandFrame("DATAOFF", idcode)
    cli.sendData(cmd_off.fullFrameBytes)


def turnDataOn(cli, idcode):
    """Send command to turn on real-time data.

    Parameters
    ----------
    cli: Client
        Client connection to data source.
    idcode: int
        Frame ID of data source.
    """
    cmd_on = CommandFrame("DATAON", idcode)
    cli.sendData(cmd_on.fullFrameBytes)


def requestConfigFrame2(cli, idcode):
    """Send command to request config frame 2.

    Parameters
    ----------
    cli: Client
        Client connection to data source.
    idcode: int
        Frame ID of data source.
    """
    cmd_config_2 = CommandFrame("CONFIG2", idcode)
    cli.sendData(cmd_config_2.fullFrameBytes)


def readConfigFrame2(cli, debug=False):
    """Retrieve and return config frame 2 from PMU or PDC.

    Parameters
    ----------
    cli: Client
        Client connection to data source.
    debug: bool
        Print debug statements.

    Returns
    -------
    : ConfigFrame
        Note: If no answer at all returns Fasle, if answer is wrong
        returns None.
    """
    leading_byte = cli.readSample(1)

    if not leading_byte:  # can't get sample at all
        return None

    if leading_byte[0] != 170:  # wrong synchronization word
        return None

    sample = leading_byte + cli.readSample(3)
    if (sample[1] & 112) != 48:  # wrong frame type
        return None

    config_frame = ConfigFrame(bytesToHexStr(sample), debug)
    exp_size = config_frame.framesize
    sample = cli.readSample(exp_size - 4)

    buf = config_frame.frame + bytesToHexStr(sample).upper()

    size = 2 * int(buf[4:8], 16)
    while len(buf) < size:
        sample = cli.readSample(1)
        buf = buf + bytesToHexStr(sample).upper()

    config_frame.frame = buf
    config_frame.finishParsing()

    return config_frame


def getDataSample(rcvr):
    """Get a data sample regardless of TCP or UDP connection.

    Parameters
    ----------
    rcvr: Object
        Object used for receiving data frames: Client or Server.

    Returns
    -------
    : str
        Data frame in hex string format.
    """

    full_hex_str = ""

    if type(rcvr) == "client":
        intro_hex_str = bytesToHexStr(rcvr.readSample(4))
        len_to_read = int(intro_hex_str[5:], 16)
        rest_hex_str = bytesToHexStr(rcvr.readSample(len_to_read))
        full_hex_str = intro_hex_str + rest_hex_str

    else:
        full_hex_str = bytesToHexStr(rcvr.readSample(64000))

    return full_hex_str


def get_data_frames(data_sample, conf_frame, rest=""):
    """Return the list of data frames from data_sample and the
    unparsed rest for the next use."""

    buf = rest + data_sample

    if len(buf) < 8:
        return [], buf

    if buf[:2] != "aa":
        return [], rewind_to(buf, "aa")

    data_frames = []
    while True:
        size = 2 * int(buf[4:8], 16)

        if len(buf) < size:
            return data_frames, buf

        try:
            data_frames.append(DataFrame(buf[:size], conf_frame))
        except Exception:
            return data_frames, rewind_to(buf[2:], "aa")

        if len(buf) == size:
            return data_frames, ""

        buf = buf[size:]

    return data_frames, ""


def startDataCapture(idcode, ip, port=4712, proto="TCP", debug=False):
    """Connect to data source, request config frame, send data start
    command.

    Parameters
    ----------
    idcode: int
        Frame ID of PMU.
    ip: str
        IP address of data source.
    port: int
        Command port on data source.
    proto: str
        Use TCP or UDP.
    debug: bool
        Print debug statements.

    Returns
    -------
    : espmu.config_frame.ConfigFrame
        Config frame.
    """
    config_frame = None

    cli = Client(ip, port, proto)
    cli.setTimeout(5)
    turnDataOff(cli, idcode)
    while config_frame is None:
        requestConfigFrame2(cli, idcode)
        config_frame = readConfigFrame2(cli, debug)
    cli.stop()

    return config_frame


def getStations(config_frame):
    """Returns all station names from the config frame.

    Parameters
    ----------
    config_frame: ConfigFrame
        ConfigFrame containing stations.

    Returns
    -------
    : list
        All the station names.
    """

    return config_frame.stations


def parseSamples(data, config_frame, pmus):
    """Takes in an array of dataFrames and inserts the data into an
    array of aggregate phasors.

    Parameters
    ----------
    data: list
        All the data samples.
    config_frame: ConfigFrame
        ConfigFrame containing stations.
    pmus: list
        Phasor values.

    Returns
    -------
    : list
        All the phasor values.
    """
    num_of_samples = len(data)
    for s in range(num_of_samples):
        for p in range(len(data[s].pmus)):
            for phasor in range(len(data[s].pmus[p].phasors)):
                msec = (data[s].fracsec / config_frame.time_base.baseDecStr)
                utc_timestamp = data[s].soc.utcSec + msec
                pmus[p][phasor].addSample(
                    utc_timestamp,
                    data[s].pmus[p].phasors[phasor].mag,
                    data[s].pmus[p].phasors[phasor].rad)

    return pmus


def rewind_to(buf, pattern):
    """Trim the buffer so that it starts with the pattern."""

    i = buf.find(pattern)
    if i == -1:
        return ""
    return buf[i:]
